// *****************************************************************
// ****                      Synth CORE B                       ****
// ****                  www.tolaemon.com/scoreb                ****
// ****                       Source code                       ****
// ****                       13/11/2011                        ****
// ****                     Jordi Bartolome                     ****
// ****                                                         ****
// **** IMPORTANT:                                              ****
// **** Using parts of this code means accepting all conditions ****
// **** exposed on the web: www.tolaemon.com/scoreb             ****
// ****                                                         ****
// **** Parts of this code have been taken from the demo source ****
// **** code of EVK1100 Atmel's evaluation board				****
// *****************************************************************

#include <avr32/io.h>
#include "board.h"
#include "gpio.h"
#include "usart.h"

// scheduler includes.
#include "FreeRTOS.h"
#include "queue.h"
#include "task.h"

// synth core B includes
#include "defines.h"
#include "midi.h"
#include "com.h"

// MIDI configuration vars
static struct {
	portCHAR initialized; // true if the MIDI is initialized and running
	portCHAR device_type; // with the decive_type the MIDI is configured to
	portCHAR device_number; // the device id where the MIDI is confiruged to
	xTaskHandle midi_task_handler; // the handler to the FreeRTOS MIDI task
} midi_config;

// MIDI rx buffer
static struct {
	portSHORT bytes_received; // with the number of bytes stored into the frame
	portSHORT bytes_to_receive; // with the number of expected btyes in the MIDI message
	portCHAR  buffer[MIDI_MAX_MESSAGE]; // buffer used to store the bytes of the received MIDI message
} midi_message;

// Queues used to hold the MIDI events
static xQueueHandle x_MIDI_tx_event;// MIDI events to transmit
static xQueueHandle x_MIDI_rx_event;// received the MIDI events



// Synth Core  30-10-2011  Tolaemon
// Function: void MIDI_dequeue_received_event(){
// Function which returns the next MIDI event stored into the
// MIDI events reception queue
// Receives:
// Returns:
// By reference:
//   	MIDI_event: with the information of the received MIDI
// 		event.
// By value:
//	+1: an event has been read from the MIDI events queue
//	-1: there was no event into the MIDI queue
signed portCHAR MIDI_dequeue_received_event(t_MIDI_event * pMIDI_event){
	signed portCHAR return_value;


	if (midi_config.initialized==TRUE){

		if( xQueueReceive( x_MIDI_rx_event, pMIDI_event, 1 ) ){
			return_value=1;
		}else{
			return_value=-1;
		}//if

	}else{
		return_value=-1;
	}//if

	return return_value;

}//MIDI_dequeue_received_event



// Synth Core  9-11-2011  Tolaemon
// Function: void MIDI_enqueue_event_to_send(){
// Function which takes a MIDI event and enqueues it into the
// x_MIDI_tx_event queue ( the queue with the events to be
// transmited through the MIDI out port ).
// Receives:
//	* pMIDI_event: a pointer to the structure with the information
//  of the MIDI event to transmit
// Returns:
// By value:
//	+1: the MIDI event was succesfully enqueued
//	-1: the MIDI event could not be enqueued
signed portCHAR MIDI_enqueue_event_to_send(t_MIDI_event * pMIDI_event){
	signed portCHAR return_value;

	if (midi_config.initialized==TRUE){

		// checks the type of the channel where the MIDI task is connected to check
		// if sending MIDI events is supported
		switch (midi_config.device_type){

				case DEVICE_TYPE_COM:
					// sending MIDI events is suported in COM type devices

					// places the MIDI event in the queue of MIDI events to send
					if( xQueueSend( x_MIDI_tx_event, pMIDI_event, 0 ) != pdPASS ){
						// the message could not be enqueued
						return_value=-1;
					}else{
						return_value=1;
					}//if

					break;

				case DEVICE_TYPE_FILE:
					// sending MIDI events is NOT suported in FILE type devices
					return_value=-1;
					break;

				default:
					// invalid
					return_value=-1;
					break;
		}//switch

	}else{
		return_value=-1;
	}//if

	return return_value;

}//MIDI_enqueue_event_to_send




// Synth Core  26-10-2011  Tolaemon
// Function: void MIDI_process_received_message(){
// Function which processes the MIDI message stored into the midi
// message buffer used to parse the messages.
// Receives:
// Returns:
//	+1: if the message could be processed
//	-1: if the message could not be processed
//	-2: if the message could be processed but not enqueued into the MIDI queue
// Note:
//		a complete MIDI message must be stored into the buffer
signed portCHAR MIDI_process_received_message(){
	t_MIDI_event midi_event_aux;
	signed portCHAR return_value;
	signed portLONG sl_aux;


	// first byte is status byte and indicates the MIDI type of message
	if ((midi_message.buffer[0]>=0x80)&&(midi_message.buffer[0]<=0xEF)){

		// it is a MIDI VOICE MESSAGE, bits 0..3 are the channel , bits 4..7 are the command
		midi_event_aux.type=MIDI_VOICE_MESSAGE_TYPE;
		midi_event_aux.code=(midi_message.buffer[0]&MIDI_MASK_VOICE_MESSAGE);
		midi_event_aux.channel=(midi_message.buffer[0]&MIDI_MASK_CHANNEL);

		// the meaning of the data bytes will be different depending on the command codes
		switch (midi_event_aux.code){
			case MIDI_COD_NOTE_OFF:
				midi_event_aux.value_1=(signed portLONG)midi_message.buffer[1]; // released note number
				midi_event_aux.value_2=(signed portLONG)midi_message.buffer[2]; // released note speed
				return_value=1;
				break;
			case MIDI_COD_NOTE_ON:
				midi_event_aux.value_1=(signed portLONG)midi_message.buffer[1]; // pressed note number
				midi_event_aux.value_2=(signed portLONG)midi_message.buffer[2]; // pressed note speed
				return_value=1;
				break;
			case MIDI_COD_AFTER_TOUCH:
				midi_event_aux.value_1=(signed portLONG)midi_message.buffer[1]; // pressure change note number
				midi_event_aux.value_2=(signed portLONG)midi_message.buffer[2]; // new value of pressure of the note
				return_value=1;
				break;
			case MIDI_COD_CTRL_CHANGE:
				midi_event_aux.value_1=(signed portLONG)midi_message.buffer[1]; // number of the control we want to change its value
				midi_event_aux.value_2=(signed portLONG)midi_message.buffer[2]; // new value of the control
				return_value=1;
				break;
			case MIDI_COD_PROGR_CHANGE:
				midi_event_aux.value_1=(signed portLONG)midi_message.buffer[1]; // number of the instrument we want to set in the channel
				midi_event_aux.value_2=0;// this command uses only one value
				return_value=1;
				break;
			case MIDI_COD_CHAN_PRESURE:
				midi_event_aux.value_1=(signed portLONG)midi_message.buffer[1]; // pressure of the notes of the instrument
				midi_event_aux.value_2=0;// this command uses only one value
				return_value=1;
				break;
			case MIDI_COD_PTICH_WHEEL:
				// 7 bits of each data byte is combined into a 14 bits value: byte 1 are the 7 lowest bits and byte 2 the 7 highest
				sl_aux=(signed portLONG)(midi_message.buffer[2]);
				sl_aux=sl_aux<<7;
				sl_aux=sl_aux|(midi_message.buffer[1]);
				midi_event_aux.value_1=sl_aux; // value of the pitch wheel
				midi_event_aux.value_2=0;// this command uses only one value
				return_value=1;
				break;
			default:
				break;
		}//switch

	}else if ((midi_message.buffer[0]>=0xF0)&&(midi_message.buffer[0]<=0xF7)){

		// it is a MIDI GENERIC MESSAGE : System Common messages
		midi_event_aux.type=MIDI_SYSTEM_COMMON_MESSAGE_TYPE;
		midi_event_aux.code=midi_message.buffer[0];
		midi_event_aux.channel=0;// these messages have no channel

		// TODO process here the different types of SYSTEM COMMON MESSAGES
		return_value=-1;

	}else if ((midi_message.buffer[0]>=0xF8)&&(midi_message.buffer[0]<=0xFF)){

		// it is a MIDI GENERIC MESSAGE : Realtime messages
		midi_event_aux.type=MIDI_SYSTEM_REALTIME_MESSAGE_TYPE;
		midi_event_aux.code=midi_message.buffer[0];
		midi_event_aux.channel=0;// these messages have no channel

		// TODO process here the different types of SYSTEM REALTIME MESSAGES
		return_value=-1;

	}else{

		return_value=-1;

	}//if

	// if the message could be processed it is enqueued into the MIDI received events queue
	if (return_value==1){

		// places the MIDI event in the queue of received events
		if( xQueueSend( x_MIDI_rx_event, &midi_event_aux, 0 ) != pdPASS ){
			// the message could not be enqueued
			return -2;
		}//if

	}//if

	return return_value;

}//MIDI_process_received_message



// Synth Core  9-11-2011  Tolaemon
// Function:MIDI_transmit_event_through_COM
// Function which takes the information contained into the
// MIDI_event struct  and send its through the configured COM
// Receives:
//	* pMIDI_event: a pointer to the structure with the information
//  of the MIDI event to transmit
// Returns:
// Note:
// This procedure should only be called if the configured device
// type is COM_TYPE.
void MIDI_transmit_event_through_COM(t_MIDI_event * pMIDI_event){
	signed portCHAR  c_aux_array[16];
	portSHORT bytes_to_send;
	signed portLONG sl_aux;
    signed portCHAR result;

	// the meaning of the data bytes will be different depending on the command codes
	switch (pMIDI_event->code){
		case MIDI_COD_NOTE_OFF:
			c_aux_array[0]=MIDI_COD_NOTE_OFF|((portCHAR)(0x0F&(pMIDI_event->channel)));
			c_aux_array[1]=(portCHAR)pMIDI_event->value_1;// released note number
			c_aux_array[2]=(portCHAR)pMIDI_event->value_2;// released note speed
			bytes_to_send=3;
			break;
		case MIDI_COD_NOTE_ON:
			c_aux_array[0]=MIDI_COD_NOTE_ON|((portCHAR)(0x0F&(pMIDI_event->channel)));
			c_aux_array[1]=(portCHAR)pMIDI_event->value_1;// pressed note number
			c_aux_array[2]=(portCHAR)pMIDI_event->value_2;// pressed note speed
			bytes_to_send=3;
			break;
		case MIDI_COD_AFTER_TOUCH:
			c_aux_array[0]=MIDI_COD_AFTER_TOUCH|((portCHAR)(0x0F&(pMIDI_event->channel)));
			c_aux_array[1]=(portCHAR)pMIDI_event->value_1;// pressure change note number
			c_aux_array[2]=(portCHAR)pMIDI_event->value_2;// new value of pressure of the note
			bytes_to_send=3;
			break;
		case MIDI_COD_CTRL_CHANGE:
			c_aux_array[0]=MIDI_COD_CTRL_CHANGE|((portCHAR)(0x0F&(pMIDI_event->channel)));
			c_aux_array[1]=(portCHAR)pMIDI_event->value_1;// number of the control we want to change its value
			c_aux_array[2]=(portCHAR)pMIDI_event->value_2;// new value of the control
			bytes_to_send=3;
			break;
		case MIDI_COD_PROGR_CHANGE:
			c_aux_array[0]=MIDI_COD_CTRL_CHANGE|((portCHAR)(0x0F&(pMIDI_event->channel)));
			c_aux_array[1]=(portCHAR)pMIDI_event->value_1;// number of the instrument we want to set in the channel
			bytes_to_send=2;
			break;
		case MIDI_COD_CHAN_PRESURE:
			c_aux_array[0]=MIDI_COD_CHAN_PRESURE|((portCHAR)(0x0F&(pMIDI_event->channel)));
			c_aux_array[1]=(portCHAR)pMIDI_event->value_1;// pressure of the notes of the instrument
			bytes_to_send=2;
			break;
		case MIDI_COD_PTICH_WHEEL:
			c_aux_array[0]=MIDI_COD_PTICH_WHEEL|((portCHAR)(0x0F&(pMIDI_event->channel)));
			// 7 bits of each data byte is combined into a 14 bits value: byte 1 are the 7 lowest bits and byte 2 the 7 highest
			sl_aux=pMIDI_event->value_1;
			c_aux_array[1]=(portCHAR)(sl_aux&0x007F); // store the 7 lowest bits
			c_aux_array[2]=(portCHAR)((sl_aux>>7)&0x007F); // store higher 7 bits
			bytes_to_send=3;
			break;
		default:
			bytes_to_send=0;
			break;
	}//switch

	// transmit the programmed message if there is one
	result=1;
	sl_aux=0;
	while ((sl_aux<bytes_to_send)&&(result!=-1)){
		// send each programmed byte through the specifyed device_number
		result=COM_xSerialPutChar(midi_config.device_number,c_aux_array[sl_aux], 0 );
		sl_aux++;
	}//while

}//MIDI_transmit_event_through_COM



// Synth Core  20-10-2011  Tolaemon
// Function: void MIDI_store_byte_into_buffer(portCHAR received_byte)
// Function which stores the received chars into the buffer and checks if a full MIDI
// message has been received. If a complete MIDI message has been received it is sent
// to the function which parses it's content.
// Receives:
//	portCHAR received_byte: with the received byte we want to store into the MIDI message buffer
// Returns:
void MIDI_store_byte_into_buffer(portCHAR received_byte){


	// TODO borrame TRATAR AQUI EL TIMEOUT DE RECEPCION DE CARACTERES


	// a char has been received, so we check if it is a MIDI status byte or a normal MIDI data byte
	if ((received_byte&MIDI_MASK_IS_A_STATUS_BYTE)!=0){

			// received byte corresponds to a status byte

			if (received_byte==MIDI_COD_END_SYSEX){

				// it is a special case, the only MIDI message with not fixed length and with an end of frame STATUS BYTE
				MIDI_process_received_message();

				// once message has been processed we can reset the buffer
				midi_message.bytes_received=0;
				midi_message.bytes_to_receive=0;

			}else{

				// standard status byte ( fist byte of a MIDI message )

				if ((received_byte&MIDI_MASK_VOICE_MESSAGE)==MIDI_COD_NOTE_ON){

					midi_message.buffer[0]=received_byte;
					midi_message.bytes_received=1;
					midi_message.bytes_to_receive=MIDI_LEN_NOTE_ON;

				}else if ((received_byte&MIDI_MASK_VOICE_MESSAGE)==MIDI_COD_NOTE_OFF){

					midi_message.buffer[0]=received_byte;
					midi_message.bytes_received=1;
					midi_message.bytes_to_receive=MIDI_LEN_NOTE_OFF;

				}else if ((received_byte&MIDI_MASK_VOICE_MESSAGE)==MIDI_COD_AFTER_TOUCH){

					midi_message.buffer[0]=received_byte;
					midi_message.bytes_received=1;
					midi_message.bytes_to_receive=MIDI_LEN_AFTER_TOUCH;

				}else if ((received_byte&MIDI_MASK_VOICE_MESSAGE)==MIDI_COD_CTRL_CHANGE){

					midi_message.buffer[0]=received_byte;
					midi_message.bytes_received=1;
					midi_message.bytes_to_receive=MIDI_LEN_CTRL_CHANGE;

				}else if ((received_byte&MIDI_MASK_VOICE_MESSAGE)==MIDI_COD_PROGR_CHANGE){

					midi_message.buffer[0]=received_byte;
					midi_message.bytes_received=1;
					midi_message.bytes_to_receive=MIDI_LEN_PROGR_CHANGE;

				}else if ((received_byte&MIDI_MASK_VOICE_MESSAGE)==MIDI_COD_CHAN_PRESURE){

					midi_message.buffer[0]=received_byte;
					midi_message.bytes_received=1;
					midi_message.bytes_to_receive=MIDI_LEN_CHAN_PRESURE;

				}else if ((received_byte&MIDI_MASK_VOICE_MESSAGE)==MIDI_COD_PTICH_WHEEL){

					midi_message.buffer[0]=received_byte;
					midi_message.bytes_received=1;
					midi_message.bytes_to_receive=MIDI_LEN_PTICH_WHEEL;

				}else{

					// TODO check here and program the expected length of the other MIDI
					// messages ( System Realime Messages, Common Realtime Messages )

					// not suported MIDI STATUS byte, so we reset the buffer
					midi_message.bytes_received=0;
					midi_message.bytes_to_receive=0;

				}//if

			}//if

	}else{

		// received byte is a normal data byte

		// if a valid message is expected ( midi_message.bytes_to_receive>0 )
		// received byte is stored into the buffer
		if (midi_message.bytes_to_receive>0){

			midi_message.buffer[midi_message.bytes_received]=received_byte;
			midi_message.bytes_received++;

			// checks if the received byte is the last byte of the MIDI message
			if (midi_message.bytes_received==midi_message.bytes_to_receive){

				// a full MIDI message has been received so we process it
				MIDI_process_received_message();

				// once message has been processed we can reset the buffer
				midi_message.bytes_received=0;
				midi_message.bytes_to_receive=0;

			}//if

		}//if

	}//if

}//MIDI_store_byte_into_buffer



// Synth Core  20-10-2011  Tolaemon
// Task which manages MIDI IN and MIDI OUT messages ( when is running ). It takes the bytes
// received through the MIDI IN device ( which can be a COM or any other device ), parses
// those bytes and converts them to an EVENT which is introduced into the MIDI IN EVENTs queue.
// It also does the inverse process taking the EVENTS from the MIDI OUT EVENTs queue, and
// converting them to its MIDI equivalent message, which is then sen through the configured
// device ( tipically teh MIDI COM  ).
static void MIDI_task(void *pvParameters ){
	signed portCHAR received_char;
	t_MIDI_event MIDI_event;

	while (1){

		// checks the type of the channel where the MIDI task is connected
		// for receiving and sending the bytes of the MIDI messages
		switch (midi_config.device_type){

				case DEVICE_TYPE_COM:

					// checks if a byte has been received through the COM where MIDI is configured
					if (COM_xSerialGetChar( midi_config.device_number,&received_char,1)>0){

						// sends the received char to the function which stores the MIDI messages
						MIDI_store_byte_into_buffer((portCHAR)received_char);

					}//if

					// checks if there is a MIDI event to send through the MIDI OUT port
					if( xQueueReceive( x_MIDI_tx_event, &MIDI_event, 1 ) ){

						// sends the MIDI event through the configured COM
						MIDI_transmit_event_through_COM(&MIDI_event);

					}//if

					break;
				case DEVICE_TYPE_FILE:

					break;

				default:

					break;
		}//switch

	}//while(1)

}//MIDI_task



// Synth Core  20-10-2011  Tolaemon
// Function: void MIDI_reset()
// Function which resets MIDI port control vars. It is used when intializing
// the system to grant all vars are on "initial conditions"
// Receives:
// Returns:
// Note:
//     Call before starting the OS scheduler
void MIDI_reset(){

	midi_config.initialized=FALSE;
	midi_config.device_type=FALSE;
	midi_config.device_number=FALSE;
	midi_config.midi_task_handler=NULL;

	midi_message.bytes_received=0;
	midi_message.bytes_to_receive=0;

}// MIDI_reset



// Synth Core  20-10-2011  Tolaemon
// Function: signed portCHAR MIDI_init(portCHAR com_number){
// Function wich initializes the MIDI comunications throguh the specified device
// Receives:
//    portUCHAR device_type: indicating which kind of device is going to be used to read MIDI
//    portUCHAR device_number: with the identifying number of the device used to read MIDI
// Returns:
// 	 1: if the MIDI has been properly initialized
//  -1:if the MIDI could not be initialized
// Note:
//   MIDI_reset(...) should have been called at the application start up
signed portCHAR MIDI_init(portCHAR device_type, portCHAR device_number){
	signed portCHAR return_value=1;

	// configure MIDI to work with the specified device
	switch (device_type){

		// MIDI messages are received and sent through a COM port
		case DEVICE_TYPE_COM:

			midi_config.initialized=TRUE;
			midi_config.device_type=device_type;
			midi_config.device_number=device_number;

			midi_message.bytes_received=0;
			midi_message.bytes_to_receive=0;

			// creates the queues used to hold the received MIDI events and the MIDI events to send
			x_MIDI_tx_event = xQueueCreate( MIDI_QUEUE_LENGTH, ( unsigned portBASE_TYPE ) sizeof( t_MIDI_event) );;
			x_MIDI_rx_event  = xQueueCreate( MIDI_QUEUE_LENGTH, ( unsigned portBASE_TYPE ) sizeof( t_MIDI_event ) );;

			// create the midi task to start processing the bytes
			xTaskCreate( MIDI_task ,  (const signed portCHAR *)"MIDI task" , MIDI_TASK_STACK_SIZE ,  NULL ,  MIDI_TASK_PRIORITY ,&midi_config.midi_task_handler);

			return_value = 1;
			break;

		// MIDI messages are received from a FILE
		case DEVICE_TYPE_FILE:
			// NOT IMPLEMENTED YET

			midi_message.bytes_received=0;
			midi_message.bytes_to_receive=0;

			midi_config.initialized=FALSE;
			return_value = -1;
			break;

		default:
			midi_config.initialized=FALSE;
			return_value = -1;
			break;

	}//switch

	return return_value;

}//MIDI_init



// Synth Core  20-10-2011  Tolaemon
// Function: MIDI_close
// Function which closes the MIDI port
// Receives:
// Returns:
// Note:
//   MIDI_reset(...) should have been called at the application start up
void MIDI_close(){

	if (midi_config.initialized==TRUE){

		// as we are not going to use MIDI, MIDI midi task is deleted
		vTaskDelete( midi_config.midi_task_handler );

		// release resources that we are not going to use
		vQueueDelete(x_MIDI_tx_event);
		vQueueDelete(x_MIDI_rx_event);

		midi_config.initialized=FALSE;
		midi_config.device_type=FALSE;
		midi_config.device_number=FALSE;
		midi_config.midi_task_handler=NULL;

	}//if

}//MIDI_close

